
# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: //:liga

from setuptools import setup

setup(**{
    'install_requires': (
        'fsspec==2022.1.0',
        'jsonschema',
        'numpy',
        'pandas',
        'py4j',
        'pyarrow>=6.0',
        'pyspark==3.2.1',
        'pyyaml',
        'requests',
    ),
    'name': 'liga',
    'namespace_packages': (
    ),
    'package_data': {
    },
    'packages': (
        'liga',
        'liga.internal',
        'liga.numpy',
        'liga.registry',
        'liga.spark',
        'liga.spark.types',
    ),
    'python_requires': '<3.10,>=3.8',
    'version': '0.3.0.dev0',
})
